#
# Copyright (c) 2002, 2003, 2004, 2005 Sebastian Stark
# Copyright (c) 2019-2021 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Interactive mode for tentakel."""

import cmd

from . import remote

try:
    import readline  # noqa
except ImportError:
    pass


class TentakelShell(cmd.Cmd):
    def __init__(self, conf, group_name):
        super().__init__()
        self.doc_header = "commands (type help <topic>):"
        self.ruler = ""
        self.group_name = group_name
        self.prompt = f"tentakel({group_name})> "
        self.conf = conf
        self.dests = remote.RemoteCollator(conf, group_name)

    def __del__(self):
        self.dests.join_all()

    def emptyline(self):
        pass

    def postcmd(self, stop, rest):
        self.prompt = f"tentakel({self.group_name})> "
        return stop

    def do_exec(self, cmd):
        """exec <cmd>: applies <cmd> to the current group."""

        if not cmd:
            print("empty command")
            return
        self.dests.exec_all(cmd)
        self.dests.display_all()

    def do_conf(self, rest):
        """conf: interactively edit current configuration"""

        self.conf.edit()
        self.dests.use_conf(self.conf, self.group_name)

    def do_use(self, rest):
        """use <groupname>: use the specified group."""

        if rest:
            self.group_name = rest
            self.dests.use_conf(self.conf, self.group_name)

    def do_groups(self, rest):
        """groups: list available groups"""

        print("\n".join(self.conf.get_groups()))

    def do_hosts(self, rest):
        """hosts: list of affected hosts"""

        print("\n".join(self.dests.get_destinations()))

    def do_quit(self, rest):
        """quit or ctrl-d: quit program."""

        self.dests.join_all()
        return 1

    def default(self, rest):
        if rest == "EOF":
            print()
            self.dests.join_all()
            return 1
        else:
            print("unknown command")

    def help_help(self):
        print("help <something>: show usage of <something>")



#!/usr/bin/env python

# Copyright (c) 2002, 2003, 2004, 2005 Sebastian Stark
# Copyright (c) 2011, 2019-2021 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Distributed command execution

Usage: tentakel [ options ] [ command ]
 -c file        Use file as config file
 -g group       Select group
 -l             Print list of available groups
 -h             Display this help text
 -v             Display version information
 command        Remote command. Interactive mode if not specified

See tentakel(1) for more information
"""

import getopt
import os
import sys

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata  # type: ignore

from tentakel import config, error, remote, shell


def main():
    group_name = "default"
    flag_listgroups = 0
    override_config = ""

    try:
        opts, args = getopt.getopt(sys.argv[1:], "g:hlvc:D")
    except getopt.GetoptError:
        print_help()
        error.err("parameter error")

    for o, v in opts:
        if o == "-h":
            print_help()
            sys.exit(0)
        if o == "-v":
            print(get_version())
            sys.exit(0)
        if o == "-D":
            print_info()
            sys.exit(0)

        if o == "-g":
            group_name = v
        if o == "-c":
            override_config = v
        if o == "-l":
            flag_listgroups = 1

    command = " ".join(args)

    # check wether the user has chosen a specific configuration file
    # on the command line
    config_file = None
    if override_config:
        if os.path.isfile(override_config):
            config_file = override_config
        else:
            error.err(f"no such file: '{override_config}'")
    else:
        # look for configuration files from default locations
        configs = [
            os.path.join(config.__user_dir, "tentakel.conf"),
            "/etc/tentakel.conf",
        ]
        for c in configs:
            if os.path.isfile(c):
                config_file = c
                break

    if config_file is None:
        error.err("no configuration file found")
        sys.exit(1)

    # load configuration
    conf = config.ConfigBase()
    with open(config_file) as f:
        conf.load(f)

    # process -g parameter
    if flag_listgroups:
        print("available groups:")
        for g in conf.get_groups():
            sys.stdout.write(g + " ")
        print()
        sys.exit(0)

    # batch mode: execute command
    if command:
        collator = remote.RemoteCollator(conf, group_name)
        collator.exec_all(command)
        collator.display_all()
        collator.join_all()
    else:
        # interactive mode: open shell
        sh = shell.TentakelShell(conf, group_name)
        sh.cmdloop(intro="interactive mode")


def print_help():
    print(__doc__)


def print_info():
    print(f"path: {__file__}")
    print(f"version: {get_version()}")


def get_version():
    return metadata.version("tentakel")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2002, 2003, 2004 Sebastian Stark
# Copyright (c) 2019-2021 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import subprocess
import time

from tentakel.remote import RemoteCommand, register_remote_command_plugin


class SSHRemoteCommand(RemoteCommand):
    """SSH remote execution class."""

    ssh_path: str
    user: str

    def __init__(self, destination, params):
        self.ssh_path = params["ssh_path"]
        self.user = params["user"]
        super().__init__(destination, params)

    def _rexec(self, command: str) -> tuple[int, str]:
        s = f'{self.ssh_path} {self.user}@{self.destination} "{command}"'
        t1 = time.time()
        status, output = subprocess.getstatusoutput(s)
        self.duration = time.time() - t1
        # shift 8 bits right to strip signal number from status
        return (status >> 8, output)


register_remote_command_plugin("ssh", SSHRemoteCommand)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tentakel', 'tentakel.plugins']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['tentakel = tentakel.main:main']}

setup_kwargs = {
    'name': 'tentakel',
    'version': '4.0.0',
    'description': 'distributed command execution',
    'long_description': 'About\n=====\n\nTentakel is a program for executing the same command on many hosts in parallel\nusing various remote methods. It can make use of several sets of hosts that are\ndefined in a configuration file as groups.\n\nIt also supports an interactive mode that can be used for repeated commands.\n\nThe author uses tentakel to simultaneously install software on many\nworkstations or ask and set parameters on a linux compute cluster.  With the\npower of format strings tentakel can also be used for monitoring purposes.\n\nSupported remote methods are ssh(1) and rsh(1).  Both must be configured to\nallow for password-less logins. Password-protected keyfiles for ssh can be\nused with ssh-agent(1).\n\nA plugin mechanism allows users to implement their own remote methods in\naddition to the builtin ones.\n\nFor more information on available options please refer to the manpage\ntentakel(1).\n\nThe project homepage is: <https://github.com/sfermigier/tentakel>\n\nThe current maintainer is:\n\n- Stefane Fermigier <sf@fermigier.com>\n\nThe original authors were:\n\n- Sebastian Stark <cran@users.sourceforge.net>\n- Marlon Berlin <imaginat@users.sourceforge.net>\n\nThis software contains the Toy Parser Generator (tpg.py)\nwritten by Christophe Delord.\n',
    'author': 'Sebastian Stark, Marlon Berlin',
    'author_email': 'cran@users.sourceforge.net, imaginat@users.sourceforge.net',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sfermigier/tentakel',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Copyright (c) 2002, 2003, 2004, 2005 Sebastian Stark
# Copyright (c) 2019-2021 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Remote command handling module.

Provides the two classes which are most important in tentakel:

  - RemoteCommand
    A basic class which needs to be subclassed by plugins. A RemoteCommand
    descendant is a single thread which runs as long as tentakel is running.
    It checks continuously if there is a command to execute by fetching it from
    its commandQueue. The command is then executed by the _rexec method and
    the result is inserted into the resultQueue.
    Subclasses must define the _rexec method in order to be useful.

  - RemoteCollator
    Container used to create and control RemoteCommand instances.
    It is also responsible for outputting the results.
"""

from __future__ import annotations

import queue
import sys
import threading
from abc import ABCMeta, abstractmethod

from . import error, tpg


class FormatString(tpg.Parser):
    r"""

    token escape  : '\\[\\nt]'  str ;
    token fmtchar  : '%[%dost]'  str ;
    token char  : '.'    str ;

    START/e -> FORMAT/e ;

    FORMAT/f ->
          $ f = ""
      ( escape/e  $ f = f + self.get_escape(e)
      | fmtchar/fc  $ f = f + self.get_special_char(fc)
      | char/c  $ f = f + c
      )*
    ;
    """

    def __init__(self):
        super().__init__()
        self.map = {r"%%": "%"}

    @property
    def format_map(self):
        """Current format character mapping"""
        return self.map

    @format_map.setter
    def format_map(self, user_map):
        self.map.update(user_map)

    def get_escape(self, e):
        """Return a dictionary which maps escape strings to literals."""
        return {r"\\": r"\\", r"\n": "\n", r"\t": "\t"}[e]

    def get_special_char(self, s):
        return self.format_map[s]


class RemoteCommand(threading.Thread, metaclass=ABCMeta):
    """Generic remote execution class.

    Specific remote command classes should inherit from this class
    and define a _rexec() method that executes the command and
    puts the result into the result queue by calling putResult().

    The __init__ method can be overridden if special processing of
    the params parameter should be done. In that case,
    RemoteCommand.__init__(self, destination, params) should
    be called at the end of __init__.

    The _rexec() method should measure the time it needs to run and
    set duration accordingly.
    """

    # auxiliary queue that holds references to objects that
    # have results ready
    finished_objects: queue.Queue = queue.Queue()

    def __init__(self, destination, params):
        super().__init__()
        self.duration = 0.0
        self.destination = destination

        self._command_queue = queue.Queue()
        self._result_queue = queue.Queue()
        self._command_timeout = 0.3
        self._sleep_period = 0.8
        self._stopevent = threading.Event()
        # In the end this will be the maxparallel value of the _outermost_ group
        # instead of the innermost, unlike all other parameters. Although
        # it is indeed predictable behaviour (because tentakel.config returns all
        # hosts in a group before it returns the subgroups hosts) it is not nice
        # and therefore a TODO
        self.__maxparallel = int(params["maxparallel"])

        # Create the semaphore as a class attribute only once, and only when needed
        if "slot" not in self.__class__.__dict__ and not self.__maxparallel <= 0:
            self.__class__.slot = threading.BoundedSemaphore(self.__maxparallel)

        self.start()

    @abstractmethod
    def _rexec(self, command):
        pass

    def execute(self, command: str):
        """Execute a command in this thread."""
        self._command_queue.put_nowait(command)

    def put_result(self, result):
        """Push result onto the result queue."""
        self._result_queue.put(result)
        self.__class__.finished_objects.put(self)

    def get_result(self):
        """Return result from result queue."""
        return self._result_queue.get()

    def run(self):
        while not self._stopevent.isSet():
            try:
                command = self._command_queue.get(timeout=self._command_timeout)
                if self.__maxparallel > 0:
                    self.slot.acquire()
                result = self._rexec(command)
                if self.__maxparallel > 0:
                    self.slot.release()
                self.put_result(result)
            except queue.Empty:
                pass
            self._stopevent.wait(self._sleep_period)

    def join(self, timeout=None):
        """Stop the thread."""
        self._stopevent.set()
        threading.Thread.join(self, self._command_timeout)


def remote_command_factory(destination, params):
    """Depending in the method, instantiate a corresponding RemoteCommand
    derived object and return it."""

    method = params["method"]
    try:
        cls = _remote_command_plugins[method]
        return cls(destination, params)
    except KeyError:
        error.err(f'Method not implemented: "{method}"')


class RemoteCollator:
    """This class is meant to hold RemoteCommand instances each of which
    implements a specific way too execute a command on a remote host."""

    def __init__(self, conf, group_name):
        self.remote_objects = []
        self.use_conf(conf, group_name)
        self.formatter = FormatString()

    def clear(self):
        """Empty the list of contained remoteobjects after stopping them."""
        for obj in self.remote_objects:
            obj.join()
        self.remote_objects = []

    def use_conf(self, conf, group_name):
        """Load the specified group from configuration object conf and add
        RemoteCommand objects for each contained host."""
        # FIXME: not sure what this does
        save = self
        self.clear()
        try:
            for destination, params in conf.get_group_members(group_name):
                self.add(remote_command_factory(destination, params))
                self.format = conf.get_param("format", group=group_name)
        except KeyError:
            self = save
            error.warn(f"unknown group: '{group_name}'")

    def get_destinations(self):
        """Return expanded list of hosts."""
        return [x.destination for x in self.remote_objects]

    def add(self, obj: RemoteCommand):
        """Add a RemoteObject."""
        assert isinstance(obj, RemoteCommand)
        self.remote_objects.append(obj)

    def remove(self, obj):
        """Remove a RemoteObject."""
        self.remote_objects.remove(obj)

    def expand_format(self, map=None):
        """Apply a format mapping to the format string.

        Outputs the format with formatting expressions replaced by
        values taken from map. The map must contain translations for
        the formatting expressions. For example:

          map = { r"%d": "something" }
        """

        self.formatter.format_map = map
        return self.formatter(self.format)

    def exec_all(self, command: str):
        """Execute command on all remote objects."""

        for obj in self.remote_objects:
            obj.execute(command)

    def join_all(self):
        """Join the running threads for all remote objects."""

        for obj in self.remote_objects:
            obj.join()

    def display_all(self):
        """Display the next pending result for every remote object."""

        display_count = len(self.remote_objects)
        while display_count > 0:
            obj = RemoteCommand.finished_objects.get()
            display_count -= 1
            status, output = obj.get_result()
            result_map = {
                "%d": obj.destination,
                "%t": str(round(obj.duration, 2)),
                "%o": output,
                "%s": str(status),
            }
            sys.stdout.write(self.expand_format(result_map))

        assert RemoteCommand.finished_objects.qsize() == 0


_remote_command_plugins = {}


def register_remote_command_plugin(method: str, cls: type[RemoteCommand]):
    """Needs to be imported and executed by remote command plugins."""
    assert issubclass(cls, RemoteCommand)

    _remote_command_plugins[method] = cls


# Don't remove / don't move
from .plugins import *  # noqa



#
# Copyright (c) 2002, 2003, 2004, 2005 Sebastian Stark
# Copyright (c) 2019-2021 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Error and user notification for tentakel."""

import sys


def err(errstring):
    sys.stderr.write(f"tentakel error: {errstring}\n")
    sys.exit(1)


def warn(warnstring):
    sys.stderr.write(f"tentakel: {warnstring}\n")



#
# Copyright (c) 2002, 2003, 2004, 2005 Sebastian Stark
# Copyright (c) 2011, 2019-2023 Stefane Fermigier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Configuration tree for tentakel.

Provides a ConfigBase class that is initialized from a configuration file.

Example:

  c = config.ConfigBase()
  f = open("tentakel.conf")
  c.load(f)
  f.close()
  availableGroups = c.getGroups()

or, if you want all hosts and their parameters with expanded sublists:

  hosts = c.getGroupMembers("mygroup")

In the latter case, only "host" objects are returned in a list of tuples,
where each tuple contains the name of the host and a complete set of
parameters, taken from its nearest enclosing group.
"""

from __future__ import annotations

import os
import pwd
import re
import sys
import tempfile

from . import error, tpg

PARAMS = {
    "ssh_path": "/usr/bin/ssh",
    "rsh_path": "/usr/bin/rsh",
    "method": "ssh",
    "maxparallel": "0",
    "user": pwd.getpwuid(os.geteuid())[0],
    "format": r"### %d(stat: %s, dur(s): %t):\n%o\n",
}

METHODS = ["ssh", "rsh"]

__user_dir = os.path.join(os.environ["HOME"], ".tentakel")
__user_plugin_dir = os.path.join(__user_dir, "plugins")


class TConf(tpg.Parser):
    __doc__ = r"""

    set lexer = ContextSensitiveLexer

    token keyword  : '{keywords}'  str ;
    token eq       : '='      str ;
    token word     : '\w+'      str ;
    token vchar    : '""|[^"]'    str ;
    token hitem    : '\+[-\w\.:]+'    str ;
    token litem    : '@\w+'    str ;

    separator spaces  : '\s+' ;

    START/e ->      $ e = {{"groups": {{}}, "settings": PARAMS}}
    (  SETTING/s    $ e["settings"].update(s)
      | GROUP/g     $ e["groups"][g["name"]] = g
      | COMMENT
    )*
    ;

    COMMENT ->   @start '\s*#.*' @end
    ;

    SETTING/s ->  'set'  $ s = {{}}
      PARAM/<p,v>        $ s[p] = v
    ;

    PARAM/<p,v> ->  keyword/p eq
      '"'
      @start (vchar)* @end  $ t = self.extract(start, end)
      '"'                   $ v = re.sub('""', '"', t)
    ;

    GROUP/g -> 'group'  $ g = ConfigGroup()
      GROUPNAME/n       $ g["name"] = n
      GROUPSPEC/s       $ g.update(s)
      MEMBERS/l         $ g.update(l)
    ;

    GROUPNAME/n -> word/n ;

    GROUPSPEC/s ->        $ s = {{}}
      '\('
        ( PARAM/<p,v>     $ s[p] = v
        )?
        ( ',' PARAM/<p,v> $ s[p] = v
        )*
      '\)'
    ;

    MEMBERS/l ->  $ l = {{"hosts": [], "lists": []}}
      ( hitem/i   $ l["hosts"].append(i[1:])
      | litem/i   $ l["lists"].append(i[1:])
      | COMMENT
      )*
    ;
    """.format(
        keywords="|".join(PARAMS.keys())
    )


class ConfigGroup(dict):
    """Store group info."""

    def __init__(self):
        super().__init__()
        self["name"] = ""
        # create keys that are also available globally, but with default values stripped
        p = dict(list(zip(list(PARAMS.keys()), [""] * len(PARAMS))))
        self.update(p)
        self["hosts"] = []
        self["lists"] = []

    def __str__(self):
        groups = []
        for param in PARAMS.keys():
            if self[param]:
                groups.append('{}="{}"'.format(param, re.sub('"', '""', self[param])))
        return f"group {self['name']} ({', '.join(groups)})"


class ConfigBase(dict):
    """Store all configuration parameters.

    This class is used to hold a specific configuration state in a special
    tree that's built out of dictionaries and lists. Single parameters can
    be changed or asked for their values. The whole configuration can be
    changed by using the parse method on a string that contains configuration
    directives in a format suitable for tentakel. Alternatively the load
    method can be used directly on a file.

    The configuration can be written to a file with the dump method.
    """

    def __init__(self):
        super().__init__()
        self["groups"] = {}
        self["settings"] = PARAMS

    def parse(self, txt):
        """Parse a string containing configuration directives into the
        configuration tree."""
        tp = TConf()
        self.update(tp(txt))

    def load(self, file):
        """Load configuration from file."""

        try:
            self.parse("".join(file.readlines()))
        except tpg.SyntacticError as excerr:  # pragma: nocover
            error.warn(f"in {file.name}: {excerr.msg}")
        except OSError:  # pragma: nocover
            error.err(f"could not read from file: '{file.name}'")

    def dump(self, file):
        """Save configuration to file."""

        comment = [
            "#\n",
            "# CURRENT CONFIGURATION\n",
            "#\n",
            "# You can change the configuration for the current session here.\n",
            "# Those changes will be lost after you quit tentakel.\n",
            "# No configuration file will be changed.\n",
            "#\n",
        ]

        try:
            file.writelines(comment)
            file.writelines(str(self))
        except OSError:  # pragma: nocover
            error.err(f"could not write to file: '{file.name}'")

    def edit(self):
        """Interactively edit configuration."""

        tempedit = tempfile.NamedTemporaryFile()
        try:
            self.dump(tempedit)
            tempedit.seek(0, 0)
            editor = os.getenv("VISUAL") or os.getenv("EDITOR") or "vi"
            os.spawnvp(os.P_WAIT, editor, [editor, tempedit.name])
            self.load(tempedit)
        finally:
            tempedit.close()

    def __str__(self):
        """Pretty print configuration."""

        out = ""
        settings = self["settings"]
        for s_param, s_value in settings.items():
            if s_value:
                out = f'{out}set {s_param}="{s_value}"\n'
        out += "\n"
        groups = self["groups"]
        for group_name, group_obj in groups.items():
            out = out + str(group_obj) + "\n"
            for list in groups[group_name]["lists"]:
                out = out + "\t@" + list + "\n"
            for host in groups[group_name]["hosts"]:
                out = out + "\t+" + host + "\n"
            out += "\n"
        return out

    def get_groups(self) -> list[str]:
        """Return list of all group names."""

        return list(self["groups"].keys())

    def _get_group(self, group_name):
        """Return group specific configuration for group_name."""

        return self["groups"][group_name]

    def get_group_members(self, group_name):
        """Return list of group_name members with sub lists expanded
        recursively."""

        group = self._get_group(group_name)
        out = [(x, self.get_group_params(group_name)) for x in group["hosts"]]
        for list in group["lists"]:
            try:
                out += self.get_group_members(list)
            except (KeyError, RuntimeError):  # pragma: nocover
                if sys.exc_info()[0] == KeyError:
                    error.warn(f"in group '{group_name}': no such group '{list}'")
                if sys.exc_info()[0] == RuntimeError:
                    error.err("runtime error: possible loop in configuration file")
        return out

    def get_param(self, param: str, group=None):
        """Return the value for param.

        If group is specified, return the groups local value for param.
        If the group has no local value or group=None or group does not
        exist, return the global value for param.

        If param is not a valid parameter identifier, return None
        """

        if param not in PARAMS.keys():  # pragma: nocover
            error.warn(f"invalid parameter: '{param}'")
            return None
        else:
            try:
                val = self._get_group(group)[param]
                if val == "":
                    return self["settings"][param]
                else:
                    return val
            except KeyError:
                return self["settings"][param]

    def get_group_params(self, group_name):
        """Return complete configuration for the group group_name."""

        return {k: self.get_param(k, group_name) for k in PARAMS.keys()}
